"""Run a simple test of the Cusser class."""


import curses
import sys
from functools import reduce

from ._misc import (
    _SUPPORTED_ATTRIBUTE_TAGS,
    _SUPPORTED_COLOR_TAGS,
    _app,
    _clear_line,
    _clear_screen,
    _move,
    _step,
)

# TODO: this is just for show, can be improved A LOT
if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(
            f"usage: {sys.argv[0]} <example>, where <example> is one of: "
            "'attributes', 'colors', 'clear', 'cursor'"
        )
        sys.exit(1)

    MESSAGE = "The quick brown fox jumps over the lazy dog"

    if sys.argv[1] == "attributes":
        text = reduce(
            lambda acc, attribute: acc + attribute(MESSAGE) + "\n",
            _SUPPORTED_ATTRIBUTE_TAGS,
            "",
        )
    elif sys.argv[1] == "colors":
        text = reduce(
            lambda acc, color: acc + color(MESSAGE) + "\n", _SUPPORTED_COLOR_TAGS, ""
        )
    elif sys.argv[1] == "clear":
        text = f"{MESSAGE}{_clear_screen}Screen cleared!\n{MESSAGE}{_clear_line}"
    elif sys.argv[1] == "cursor":
        text = f"{MESSAGE}{_move()}{_step(1, 1)}{MESSAGE}{_move(3, 3)}{MESSAGE}"
    else:
        raise ValueError(
            f"unknown example: {sys.argv[1]}, must be one of: "
            "'attributes', 'colors', 'clear', 'cursor'"
        )

    curses.wrapper(lambda stdscr: _app(stdscr, text))



"""Utilities for managing colors and color pairs in curses."""


from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable, Iterator, Mapping, MutableSet, Optional, Text, Union

import ochre


# TODO: move this to ochre
@dataclass
class ColorPair:
    """A color pair of foreground and background colors."""

    foreground: Optional[ochre.Color] = None
    background: Optional[ochre.Color] = None


# TODO: maybe this should become hashes in ochre?
def encode(
    value: Optional[ochre.Color] | ColorPair,
) -> Optional[Text] | tuple[Optional[Text], Optional[Text]]:
    """Encode a color or color pair into optional strings to use as dictionary key."""
    if value is None:
        return None

    if isinstance(value, ochre.Color):
        return hex(value)

    if isinstance(value, ColorPair):
        return (
            hex(value.foreground) if value.foreground else None,
            hex(value.background) if value.background else None,
        )

    raise TypeError(f"Unsupported type: {type(value)}")


@dataclass
class ColorManager(
    Mapping[Union[Optional[ochre.Color], ColorPair], int],
    MutableSet[Union[Optional[ochre.Color], ColorPair]],
):
    """A class for managing curses colors and color pairs."""

    color_indices: dict[Optional[Text], int] = field(default_factory=lambda: {None: -1})
    next_color_index = 0

    pair_indices: dict[tuple[Optional[Text], Optional[Text]], int] = field(
        default_factory=dict
    )
    next_pair_index = 0

    on_add_color: Optional[Callable[[ochre.Color, ColorManager], None]] = None
    on_add_pair: Optional[Callable[[ColorPair, ColorManager], None]] = None

    current_pair: ColorPair = ColorPair()

    @property
    def foreground(self) -> ochre.Color:
        """Return the current foreground color."""
        return self.current_pair.foreground

    @foreground.setter
    def foreground(self, color: ochre.Color) -> None:
        """Set the current foreground color."""
        self.current_pair.foreground = color
        self.add_color(color)
        self.add_pair(self.current_pair)

    @property
    def background(self) -> ochre.Color:
        """Return the current background color."""
        return self.current_pair.background

    @background.setter
    def background(self, color: ochre.Color) -> None:
        """Set the current background color."""
        self.current_pair.background = color
        self.add_color(color)
        self.add_pair(self.current_pair)

    @property
    def colors(self) -> Iterator[ochre.Color]:
        """Return all colors currently registered."""
        return map(
            lambda c: ochre.Hex(c) if c is not None else None, self.color_indices.keys()
        )

    @property
    def pairs(self) -> Iterator[ColorPair]:
        """Return all color pairs currently registered."""
        return map(
            lambda p: ColorPair(
                foreground=ochre.Hex(p[0]) if p[0] is not None else None,
                background=ochre.Hex(p[1]) if p[1] is not None else None,
            ),
            self.pair_indices.keys(),
        )

    def add_color(self, color: Optional[ochre.Color], callback: bool = True) -> None:
        """Register a color with the color manager."""
        c = encode(color)
        if c in self.color_indices:
            return

        assert c is None or isinstance(c, Text)
        self.color_indices[c] = self.next_color_index
        self.next_color_index += 1
        if callback and self.on_add_color:
            self.on_add_color(color, self)

    def add_pair(
        self, pair: ColorPair, callback: bool = True, allow_zero: bool = False
    ) -> None:
        """Register a color pair with the color manager."""
        # We want background to be added first because curses tends to use the
        # first (zeroth) color as the "unknown" color, and it is usually black.
        self.add_color(pair.background, callback=callback)
        self.add_color(pair.foreground, callback=callback)

        p = encode(pair)
        if p in self.pair_indices:
            return

        if not allow_zero and self.next_pair_index == 0:
            # TODO: this might become a warning in the future?
            raise RuntimeError("Cannot redefine color pair 0")

        assert isinstance(p, tuple)
        self.pair_indices[p] = self.next_pair_index
        self.next_pair_index += 1
        if callback and self.on_add_pair:
            self.on_add_pair(pair, self)

    def discard_color(self, color: Optional[ochre.Color]) -> None:
        """Unregister a color from the color manager."""
        c = encode(color)
        if c not in self.color_indices:
            return

        assert c is None or isinstance(c, Text)
        del self.color_indices[c]

    def discard_pair(self, pair: ColorPair) -> None:
        """Unregister a color pair from the color manager."""
        p = encode(pair)
        if p not in self.pair_indices:
            return

        assert isinstance(p, tuple)
        del self.pair_indices[p]

    def add(
        self, value: Optional[ochre.Color] | ColorPair, allow_zero: bool = False
    ) -> None:
        """Register a color or color pair with the color manager."""
        if value is None or isinstance(value, ochre.Color):
            return self.add_color(value)

        if isinstance(value, ColorPair):
            return self.add_pair(value, allow_zero=allow_zero)

        raise TypeError(f"Unsupported type: {type(value)}")

    def discard(self, value: Optional[ochre.Color] | ColorPair) -> None:
        """Unregister a color or color pair from the color manager."""
        if value is None or isinstance(value, ochre.Color):
            return self.discard_color(value)

        if isinstance(value, ColorPair):
            return self.discard_pair(value)

        raise TypeError(f"Unsupported type: {type(value)}")

    def __getitem__(self, value: Optional[ochre.Color] | ColorPair) -> int:
        """Return the index of a color or color pair."""
        if value is None or isinstance(value, ochre.Color):
            c = encode(value)

            assert c is None or isinstance(c, Text)
            return self.color_indices[c]

        if isinstance(value, ColorPair):
            p = encode(value)

            assert isinstance(p, tuple)
            return self.pair_indices[p]

        raise TypeError(f"Unsupported type: {type(value)}")

    def __iter__(self) -> Iterator[ochre.Color]:
        """Return an iterator over all colors currently registered."""
        return self.colors

    def __len__(self) -> int:
        """Return the number of colors currently registered."""
        return len(self.color_indices)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cusser']

package_data = \
{'': ['*']}

install_requires = \
['stransi>=0.1.0,<0.2.0']

setup_kwargs = {
    'name': 'cusser',
    'version': '0.1.1',
    'description': 'A curses wrapper that understands ANSI escape code sequences',
    'long_description': '[![PyPI](https://img.shields.io/pypi/v/cusser)](https://pypi.org/project/cusser/)\n[![Python package](https://github.com/getcuia/cusser/actions/workflows/python-package.yml/badge.svg)](https://github.com/getcuia/cusser/actions/workflows/python-package.yml)\n[![PyPI - License](https://img.shields.io/pypi/l/cusser)](https://github.com/getcuia/cusser/blob/main/LICENSE)\n\n# [cusser](https://github.com/getcuia/cusser#readme) 🤬\n\n<div align="center">\n    <img class="hero" src="https://github.com/getcuia/cusser/raw/main/banner.jpg" alt="cusser" width="33%" />\n</div>\n\n> A curses wrapper that understands ANSI escape code sequences\n\ncusser is a lightweight Python package for teaching\n[curses](https://docs.python.org/3/library/curses.html) how to use\n[ANSI escape code sequences](https://en.wikipedia.org/wiki/ANSI_escape_code). It\nworks by wrapping the curses standard window object and intercepting escape code\nsequences.\n\n## Features\n\n-   ♻️ Easily integrate with the\n    [standard `curses` module](https://docs.python.org/3/library/curses.html)\n-   🕹️ Use the same escape code sequences\n    [as you would with Colorama](https://github.com/tartley/colorama#recognised-ansi-sequences)\n-   🖍️ Only one dependency: [stransi](https://github.com/getcuia/stransi) (for\n    actuallly parsing escape code sequences)\n-   🐍 Python 3.8+\n\n## Installation\n\n```console\n$ pip install cusser\n```\n\n## Usage\n\n```python\nIn [1]: import curses\n\nIn [2]: from cusser import Cusser\n\nIn [3]: def app(stdscr) -> None:\n   ...:     """Start a new application."""\n   ...:     if not isinstance(stdscr, Cusser):\n   ...:         stdscr = Cusser(stdscr)\n   ...:\n   ...:     ultra_violet = (100, 83, 148)\n   ...:     x, y = 34, 12\n   ...:     stdscr.addstr(\n   ...:         f"\\033[2J\\033[{x};{y}H"\n   ...:         "\\033[1;32mHello "\n   ...:         f"\\033[;3;38;2;{\';\'.join(map(str, ultra_violet))}m"\n   ...:         "cusser"\n   ...:         "\\033[m 🤬!"\n   ...:     )\n   ...:     stdscr.refresh()\n   ...:     stdscr.getch()\n   ...:\n\nIn [4]: curses.wrapper(app)\n\n```\n\n![Screenshot](https://github.com/getcuia/cusser/raw/main/screenshot.png)\n\n## Credits\n\n[Photo](https://github.com/getcuia/cusser/raw/main/banner.jpg) by\n[Gwendal Cottin](https://unsplash.com/@gwendal?utm_source=unsplash&utm_medium=referral&utm_content=creditCopyText)\non\n[Unsplash](https://unsplash.com/?utm_source=unsplash&utm_medium=referral&utm_content=creditCopyText).\n',
    'author': 'Felipe S. S. Schneider',
    'author_email': 'schneider.felipe@posgrad.ufsc.br',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/getcuia/cusser',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Private miscellaneous utilities."""


from __future__ import annotations

import curses
from typing import Callable, Text

from stransi import Ansi

from cusser import Cusser

# TODO: we'll have a library in the future for this


def _tag(start: Text, end: Text = "\033[m") -> Callable[[Text], Ansi]:
    """Define a tag function for ANSI escape sequences."""

    def decorator(text: Text) -> Ansi:
        return Ansi(f"{start}{text}{end}")

    return decorator


_normal = _tag("\033[m", "")
_bold = _tag("\033[1m", "\033[22m")
_dim = _tag("\033[2m", "\033[22m")
_italic = _tag("\033[3m", "\033[23m")
_underline = _tag("\033[4m", "\033[24m")
_blink = _tag("\033[5m", "\033[25m")
_reverse = _tag("\033[7m", "\033[27m")
_hidden = _tag("\033[8m", "\033[28m")

_SUPPORTED_ATTRIBUTE_TAGS = (
    _normal,
    _bold,
    _dim,
    _italic,
    _underline,
    _blink,
    _reverse,
    _hidden,
)


def _colortag(start: Text) -> Callable[[Text], Ansi]:
    """Define a tag function for a color ANSI escape sequence."""
    return _tag(start, "\033[39m")


_black = _colortag("\033[90m")
_red = _colortag("\033[91m")
_green = _colortag("\033[92m")
_yellow = _colortag("\033[93m")
_blue = _colortag("\033[94m")
_magenta = _colortag("\033[95m")
_cyan = _colortag("\033[96m")
_white = _colortag("\033[97m")


_SUPPORTED_COLOR_TAGS = (_black, _red, _green, _yellow, _blue, _magenta, _cyan, _white)


def _move(y: int = 0, x: int = 0) -> Ansi:
    """Move the cursor to the given position."""
    return Ansi(f"\033[{y};{x}H")


def _step(y: int, x: int) -> Ansi:
    """Move the cursor relative to its current position."""
    res = ""

    if x > 0:
        res += f"\033[{x}C"
    elif x < 0:
        res += f"\033[{-x}D"

    if y > 0:
        res += f"\033[{y}B"
    elif y < 0:
        res += f"\033[{-y}A"

    return Ansi(res)


_clear_screen = Ansi("\033[2J")
_clear_line = Ansi("\033[2K")


def _app(
    stdscr: Cusser | curses._CursesWindow, text: Text | Callable[[], Text]
) -> None:
    """Start a new application for testing."""
    if callable(text):
        text = text()
    if not isinstance(stdscr, Cusser):
        stdscr = Cusser(stdscr)
    stdscr.addstr(text)
    stdscr.refresh()
    stdscr.getch()



"""A curses wrapper that understands ANSI escape code sequences."""


from __future__ import annotations

import curses
from dataclasses import dataclass
from typing import Text

import ochre
from stransi import Ansi, SetAttribute, SetClear, SetColor, SetCursor
from stransi.attribute import Attribute
from stransi.clear import Clear
from stransi.color import ColorRole
from stransi.cursor import CursorMove

from .color_manager import ColorManager, ColorPair

__version__ = "0.1.1"


__all__ = ["Cusser"]


def on_add_color(color: ochre.Color, manager: ColorManager) -> None:
    """Initialize a color when it is added to the color manager."""
    color = color.rgb
    return curses.init_color(
        manager[color],
        int(1000 * color.red),
        int(1000 * color.green),
        int(1000 * color.blue),
    )


def on_add_pair(pair: ColorPair, manager: ColorManager) -> None:
    """Initialize a color pair when it is added to the color manager."""
    curses.init_pair(manager[pair], manager[pair.foreground], manager[pair.background])


@dataclass
class Cusser:
    """A curses wrapper that understands ANSI escape code sequences."""

    window: curses._CursesWindow
    color_manager: ColorManager = ColorManager(
        on_add_color=on_add_color, on_add_pair=on_add_pair
    )

    _ON_ATTR_MAP = {
        Attribute.BOLD: curses.A_BOLD,
        Attribute.DIM: curses.A_DIM,
        Attribute.ITALIC: curses.A_ITALIC,
        Attribute.UNDERLINE: curses.A_UNDERLINE,
        Attribute.BLINK: curses.A_BLINK,
        Attribute.REVERSE: curses.A_REVERSE,
        Attribute.HIDDEN: curses.A_INVIS,
    }

    _OFF_ATTR_MAP = {
        Attribute.NEITHER_BOLD_NOR_DIM: curses.A_BOLD | curses.A_DIM,
        Attribute.NOT_ITALIC: curses.A_ITALIC,
        Attribute.NOT_UNDERLINE: curses.A_UNDERLINE,
        Attribute.NOT_BLINK: curses.A_BLINK,
        Attribute.NOT_REVERSE: curses.A_REVERSE,
        Attribute.NOT_HIDDEN: curses.A_INVIS,
    }

    def __post_init__(self):
        """
        Initialize the color manager.

        We assume the terminal actually supports colors.
        """
        curses.start_color()
        curses.use_default_colors()

        # We have to define the zero color pair here, otherwise we'll get a nasty
        # error later.
        self.color_manager.add_pair(
            ColorPair(ochre.WebColor("white"), ochre.WebColor("black")), allow_zero=True
        )
        # TODO: initialize the same colors curses does by default, see defaults in
        # [curses.start_color](https://docs.python.org/3/library/curses.html#curses.start_color)
        # (check their RGB values with
        # [curses.color_content](https://docs.python.org/3/library/curses.html#curses.color_content)).

    def addstr(self, text: Text) -> None:
        """Add a string to the window, interpreting ANSI escape codes."""
        for instruction in Ansi(text).instructions():
            if isinstance(instruction, Text):
                self.window.addstr(instruction)
            elif isinstance(instruction, SetAttribute):
                self._set_attribute(instruction.attribute)
            elif isinstance(instruction, SetColor):
                self._set_color(instruction.role, instruction.color)
            elif isinstance(instruction, SetClear):
                self._set_clear(instruction.region)
            elif isinstance(instruction, SetCursor):
                self._set_cursor(instruction.move)
            else:
                raise NotImplementedError(instruction)

    def _set_attribute(self, attribute: Attribute) -> None:
        """Set the current attribute."""
        if attribute == Attribute.NORMAL:
            self._set_color(ColorRole.FOREGROUND, None)
            self._set_color(ColorRole.BACKGROUND, None)
            return self.window.attrset(curses.A_NORMAL)

        if attribute in self._ON_ATTR_MAP:
            return self.window.attron(self._ON_ATTR_MAP[attribute])

        if attribute in self._OFF_ATTR_MAP:
            return self.window.attroff(self._OFF_ATTR_MAP[attribute])

        raise ValueError(f"Unsupported attribute: {attribute}")

    def _set_color(self, role: ColorRole, color: ochre.Color) -> None:
        """Set the current color."""
        if role == ColorRole.FOREGROUND:
            self.color_manager.foreground = color
        elif role == ColorRole.BACKGROUND:
            self.color_manager.background = color
        else:
            raise ValueError(f"Unknown color role {role}")

        self.window.attron(
            curses.color_pair(self.color_manager[self.color_manager.current_pair])
        )

    def _set_clear(self, region: Clear) -> None:
        """Set the current clear region."""
        if region == Clear.SCREEN:
            return self.window.erase()

        if region == Clear.SCREEN_AFTER:
            return self.window.clrtobot()

        if region == Clear.LINE:
            # https://stackoverflow.com/a/20311594/4039050
            y, x = self.window.getyx()
            self.window.move(y, 0)
            return self.window.clrtoeol()
            # self.window.move(y, x)

        if region == Clear.LINE_AFTER:
            return self.window.clrtoeol()

        raise ValueError(f"Unsupported clear region {region}")

    def _set_cursor(self, move: CursorMove) -> None:
        """Set the current cursor position."""
        if not move.relative:
            y, x = move.y, move.x
        else:
            y, x = self.window.getyx()
            y += move.y
            x += move.x
        self.window.move(y, x)

    def __getattr__(self, name):
        """Forward all other calls to the underlying window."""
        return getattr(self.window, name)
